@@ -1,45 +1,44 @@
 #!/usr/bin/env python
 #-----------------------------------------------------------------------------
-#   Copyright (c) 2008, David P. D. Moss. All rights reserved.
+#   Copyright (c) 2008-2009, David P. D. Moss. All rights reserved.
 #
 #   Released under the BSD license. See the LICENSE file for details.
 #-----------------------------------------------------------------------------
 """
 network address conversion logic, constants and shared strategy objects.
 """
-import socket as _socket
 import struct as _struct
 import re as _re
+from netaddr.util import BYTES_TO_BITS as _BYTES_TO_BITS
 
-USE_IPV4_OPT=True   #: Use optimised IPv4 strategy? Default: True
-
+#   Check whether we need to use fallback code or not.
 try:
-    #   Detects a common bug in various Python implementations.
+    import socket as _socket
+    #   Check for a common bug on Windows and some other socket modules.
     _socket.inet_aton('255.255.255.255')
+    from socket import inet_aton as _inet_aton, \
+                       inet_ntoa as _inet_ntoa, \
+                       AF_INET as _AF_INET
 except:
-    USE_IPV4_OPT=False
+    from netaddr.fallback import inet_aton as _inet_aton, \
+                                 inet_ntoa as _inet_ntoa, \
+                                 AF_INET as _AF_INET
+try:
+    import socket as _socket
+    #   These might all generate exceptions on different platforms.
+    _socket.inet_pton
+    _socket.AF_INET6
+    from socket import inet_pton as _inet_pton, \
+                       inet_ntop as _inet_ntop, \
+                       AF_INET6 as _AF_INET6
+except:
+    from netaddr.fallback import inet_pton as _inet_pton, \
+                                 inet_ntop as _inet_ntop, \
+                                 AF_INET6 as _AF_INET6
 
 from netaddr import BIG_ENDIAN_PLATFORM, AT_UNSPEC, AT_INET, AT_INET6, \
                     AT_LINK, AT_EUI64, AT_NAMES, AddrFormatError
 
-#-----------------------------------------------------------------------------
-def _BYTES_TO_BITS():
-    """
-    Generates a 256 element list of 8-bit binary digit strings. List index is
-    equivalent to the bit string value.
-    """
-    lookup = []
-    bits_per_byte = range(7, -1, -1)
-    for num in range(256):
-        bits = 8*[None]
-        for i in bits_per_byte:
-            bits[i] = '01'[num&1]
-            num >>= 1
-        lookup.append(''.join(bits))
-    return lookup
-
-_BYTES_TO_BITS = _BYTES_TO_BITS()
-
 #-----------------------------------------------------------------------------
 class AddrStrategy(object):
     """Basic support for common operations performed on each address type"""
@@ -52,7 +51,7 @@ class AddrStrategy(object):
     }
 
     def __init__(self, width, word_size, word_sep, word_fmt='%x',
-                 addr_type=AT_UNSPEC, word_base=16, uppercase=False):
+                 addr_type=AT_UNSPEC, word_base=16):
         """
         Constructor.
 
@@ -73,9 +72,6 @@ class AddrStrategy(object):
 
         @param word_base: number base used to convert each word using int().
             (Default: 16)
-
-        @param uppercase: uppercase address.
-            (Default: False)
         """
 
         self.width = width
@@ -87,7 +83,6 @@ class AddrStrategy(object):
         self.word_fmt  = word_fmt
         self.word_base = word_base
         self.addr_type = addr_type
-        self.uppercase = uppercase
 
         try:
             self.name = AT_NAMES[addr_type]
@@ -98,7 +93,7 @@ class AddrStrategy(object):
         """@return: executable Python string to recreate equivalent object"""
         return "%s(%r, %r, %r, %r, %r, %r)" %  (self.__class__.__name__,
             self.width, self.word_size, self.word_sep, self.addr_type,
-            self.word_base, self.uppercase)
+            self.word_base)
 
     #-------------------------------------------------------------------------
     #   Binary methods.
@@ -130,21 +125,14 @@ class AddrStrategy(object):
         """
         @param bits: A network address in readable binary form.
 
-        @return: A network byte order integer that is equivalent to value
-            represented by network address in readable binary form.
+        @return: An unsigned integer that is equivalent to value represented
+            by network address in readable binary form.
         """
         if not self.valid_bits(bits):
             raise ValueError('%r is not a valid binary form string for ' \
                 'address type!' % bits)
 
-        word_bits = bits.split(self.word_sep)
-        if len(word_bits) != self.num_words:
-            raise ValueError('invalid number of words within binary form ' \
-                'string for address type!' % bits)
-
-        words = tuple([int(i, 2) for i in word_bits])
-
-        return self.words_to_int(words)
+        return int(bits.replace(self.word_sep, ''), 2)
 
     #-------------------------------------------------------------------------
     #   Integer methods.
@@ -152,10 +140,10 @@ class AddrStrategy(object):
 
     def valid_int(self, int_val):
         """
-        @param int_val: A network byte order integer.
+        @param int_val: An unsigned integer.
 
-        @return: C{True} if network byte order integer falls within the
-            boundaries of this address type, C{False} otherwise.
+        @return: C{True} if integer falls within the boundaries of this
+            address type, C{False} otherwise.
         """
         if not isinstance(int_val, (int, long)):
             return False
@@ -164,26 +152,26 @@ class AddrStrategy(object):
 
     def int_to_str(self, int_val):
         """
-        @param int_val: A network byte order integer.
+        @param int_val: An unsigned integer.
 
         @return: A network address in string form that is equivalent to value
-            represented by a network byte order integer.
+            represented by an unsigned integer.
         """
         words = self.int_to_words(int_val)
         tokens = [self.word_fmt % i for i in words]
         addr = self.word_sep.join(tokens)
 
-        if self.uppercase is True:
-            return addr.upper()
-
         return addr
 
-    def int_to_bits(self, int_val):
+    def int_to_bits(self, int_val, word_sep=None):
         """
-        @param int_val: A network byte order integer.
+        @param int_val: An unsigned integer.
+
+        @param word_sep: (optional) the separator to insert between words.
+            Default: None - use default separator for address type.
 
         @return: A network address in readable binary form that is equivalent
-            to value represented by a network byte order integer.
+            to value represented by an unsigned integer.
         """
         bit_words = []
 
@@ -197,11 +185,18 @@ class AddrStrategy(object):
             bits = ('0'*self.word_size+bit_str)[-self.word_size:]
             bit_words.append(bits)
 
+        if word_sep is not None:
+            #   Custom separator.
+            if not hasattr(word_sep, 'join'):
+                raise ValueError('Word separator must be a string!')
+            return word_sep.join(bit_words)
+
+        #   Default separator.
         return self.word_sep.join(bit_words)
 
     def int_to_bin(self, int_val):
         """
-        @param int_val: A network byte order integer.
+        @param int_val: An unsigned integer.
 
         @return: A network address in standard binary representation format
             that is equivalent to integer address value. Essentially a back
@@ -223,7 +218,7 @@ class AddrStrategy(object):
 
     def int_to_words(self, int_val, num_words=None, word_size=None):
         """
-        @param int_val: A network byte order integer to be split.
+        @param int_val: An unsigned integer to be divided up into words.
 
         @param num_words: (optional) number of words expected in return value
             tuple. Uses address type default if not specified.
@@ -247,16 +242,15 @@ class AddrStrategy(object):
             word = int_val & max_word_size
             words.append(int(word))
             int_val >>= word_size
-        words.reverse()
 
-        return tuple(words)
+        return tuple(reversed(words))
 
     def int_to_packed(self, int_val):
         """
         @param int_val: the integer to be packed.
 
-        @return: a packed string that is equivalent to value represented by a
-        network byte order integer.
+        @return: a packed string that is equivalent to value represented by an
+        unsigned integer.
         """
 
         words = self.int_to_words(int_val, self.num_words, self.word_size)
@@ -277,10 +271,10 @@ class AddrStrategy(object):
     def packed_to_int(self, packed_int):
         """
         @param packed_int: a packed string containing an unsigned integer.
-            Network byte order is assumed.
+            It is assumed that the string is packed in network byte order.
 
-        @return: A network byte order integer that is equivalent to value
-            of network address represented by packed binary string.
+        @return: An unsigned integer equivalent to value of network address
+            represented by packed binary string.
         """
         try:
             fmt = '>%d%s' % (self.num_words, AddrStrategy.STRUCT_FORMATS[
@@ -289,10 +283,9 @@ class AddrStrategy(object):
             raise ValueError('unsupported word size: %d!' % self.word_size)
 
         words = list(_struct.unpack(fmt, packed_int))
-        words.reverse()
 
         int_val = 0
-        for i, num in enumerate(words):
+        for i, num in enumerate(reversed(words)):
             word = num
             word = word << self.word_size * i
             int_val = int_val | word
@@ -332,8 +325,8 @@ class AddrStrategy(object):
         """
         @param addr: A network address in string form.
 
-        @return: A network byte order integer that is equivalent to value
-            represented by network address in string form.
+        @return: An unsigned integer equivalent to value represented by network
+            address in string form.
         """
         if not self.valid_str(addr):
             raise ValueError('%r is not a recognised string representation' \
@@ -373,20 +366,14 @@ class AddrStrategy(object):
         """
         @param words: A list or tuple containing integer word values.
 
-        @return: A network byte order integer that is equivalent to value
-            represented by word sequence.
+        @return: An unsigned integer that is equivalent to value represented
+            by word sequence.
         """
         if not self.valid_words(words):
             raise ValueError('%r is not a valid word list!' % words)
 
-        #   tuples have no reverse() method and reversed() is only available
-        #   in Python 2.4. Ugly but necessary.
-        if isinstance(words, tuple):
-            words = list(words)
-        words.reverse()
-
         int_val = 0
-        for i, num in enumerate(words):
+        for i, num in enumerate(reversed(words)):
             word = num
             word = word << self.word_size * i
             int_val = int_val | word
@@ -394,55 +381,13 @@ class AddrStrategy(object):
         return int_val
 
 #-----------------------------------------------------------------------------
-class IPv4StrategyStd(AddrStrategy):
-    """
-    A 'safe' L{AddrStrategy} for IPv4 addresses. Unlike L{IPv4StrategyOpt}.
-
-    It contains all methods related to IPv4 addresses that the optimised
-    version has, without the reliance on the socket or struct modules. There
-    are several cases where the use of this class are preferable when either
-    the modules mentioned do not exist on certain Python implementations or
-    contain bugs like the infamous inet_aton('255.255.255.254') bug.
-
-    All methods shared between the optimised class and this one should be
-    defined here.
-    """
+class IPv4Strategy(AddrStrategy):
+    """An L{AddrStrategy} for IPv4 address processing."""
     def __init__(self):
         """Constructor."""
-        super(IPv4StrategyStd, self).__init__(width=32, word_size=8,
+        super(IPv4Strategy, self).__init__(width=32, word_size=8,
               word_fmt='%d', word_sep='.', addr_type=AT_INET, word_base=10)
 
-    def int_to_arpa(self, int_val):
-        """
-        @param int_val: A network byte order integer.
-
-        @return: The reverse DNS lookup for an IPv4 address in network byte
-            order integer form.
-        """
-        words = ["%d" % i for i in self.int_to_words(int_val)]
-        words.reverse()
-        words.extend(['in-addr', 'arpa', ''])
-        return '.'.join(words)
-
-#-----------------------------------------------------------------------------
-class IPv4StrategyOpt(IPv4StrategyStd):
-    """
-    An optimised L{AddrStrategy} for IPv4 addresses.
-
-    It uses C{pack()} and C{unpack()} from the C{struct} module along with the
-    C{inet_ntoa()} and C{inet_aton()} from the C{socket} module great improve
-    the speed of certain operations (approx. 2.5 times faster than a standard
-    L{AddrStrategy} configured for IPv4).
-
-    However, keep in mind that these modules might not be available everywhere
-    that Python itself is. Runtimes such as Google App Engine gut the
-    C{socket} module. C{struct} is also limited to processing 32-bit integers
-    which is fine for IPv4 but isn't suitable for IPv6.
-    """
-    def __init__(self):
-        """Constructor."""
-        super(IPv4StrategyOpt, self).__init__()
-
     def valid_str(self, addr):
         """
         @param addr: An IP address in presentation (string) format.
@@ -450,16 +395,12 @@ class IPv4StrategyOpt(IPv4StrategyStd):
         @return: C{True} if network address in string form is valid for this
             address type, C{False} otherwise.
         """
+        if addr == '':
+            raise AddrFormatError('Empty strings are not supported!')
+
         try:
-            #   This call handles a lot of older IPv4 address formats that
-            #   it would be a lot of work to implement (many edge cases).
-            #   Please Note: only this optimised strategy class will have
-            #   the ability to parse these less common formats for the
-            #   moment.
-            _socket.inet_aton(addr)
-        except _socket.error:
-            return False
-        except TypeError:
+            _inet_aton(addr)
+        except:
             return False
         return True
 
@@ -467,35 +408,40 @@ class IPv4StrategyOpt(IPv4StrategyStd):
         """
         @param addr: An IPv4 dotted decimal address in string form.
 
-        @return: A network byte order integer that is equivalent to value
-            represented by the IPv4 dotted decimal address string.
+        @return: An unsigned integer that is equivalent to value represented
+            by the IPv4 dotted decimal address string.
         """
-        if not self.valid_str(addr):
-            raise ValueError('%r is not a valid IPv4 address string!' \
+        if addr == '':
+            raise AddrFormatError('Empty strings are not supported!')
+        try:
+            return _struct.unpack('>I', _inet_aton(addr))[0]
+        except:
+            raise AddrFormatError('%r is not a valid IPv4 address string!' \
                 % addr)
-        return _struct.unpack('>I', _socket.inet_aton(addr))[0]
 
     def int_to_str(self, int_val):
         """
-        @param int_val: A network byte order integer.
+        @param int_val: An unsigned integer.
 
         @return: An IPv4 dotted decimal address string that is equivalent to
-            value represented by a 32 bit integer in network byte order.
+            value represented by a 32 bit unsigned integer.
         """
-        if not self.valid_int(int_val):
-            raise ValueError('%r is not a valid 32-bit integer!' % int_val)
-        return _socket.inet_ntoa(_struct.pack('>I', int_val))
+        if self.valid_int(int_val):
+            return _inet_ntoa(_struct.pack('>I', int_val))
+        else:
+            raise ValueError('%r is not a valid 32-bit unsigned integer!' \
+                % int_val)
 
     def int_to_words(self, int_val, num_words=None, word_size=None):
         """
-        @param int_val: A network byte order integer.
+        @param int_val: An unsigned integer.
 
         @param num_words: (unused) *** interface compatibility only ***
 
         @param word_size: (unused) *** interface compatibility only ***
 
         @return: An integer word (octet) sequence that is equivalent to value
-            represented by network byte order integer.
+            represented by an unsigned integer.
         """
         if not self.valid_int(int_val):
             raise ValueError('%r is not a valid integer value supported ' \
@@ -506,28 +452,37 @@ class IPv4StrategyOpt(IPv4StrategyStd):
         """
         @param octets: A list or tuple containing integer octets.
 
-        @return: A network byte order integer that is equivalent to value
-            represented by word (octet) sequence.
+        @return: An unsigned integer that is equivalent to value represented
+            by word (octet) sequence.
         """
         if not self.valid_words(octets):
             raise ValueError('%r is not a valid octet list for an IPv4 ' \
                 'address!' % octets)
         return _struct.unpack('>I', _struct.pack('4B', *octets))[0]
 
+    def int_to_arpa(self, int_val):
+        """
+        @param int_val: An unsigned integer.
+
+        @return: The reverse DNS lookup for an IPv4 address in network byte
+            order integer form.
+        """
+        words = ["%d" % i for i in self.int_to_words(int_val)]
+        words.reverse()
+        words.extend(['in-addr', 'arpa', ''])
+        return '.'.join(words)
+
 #-----------------------------------------------------------------------------
 class IPv6Strategy(AddrStrategy):
     """
-    Implements the operations that can be performed on an Internet Protocol
-    version 6 network address in accordance with RFC 4291.
+    An L{AddrStrategy} for IPv6 address processing.
 
-    NB - This class would benefit greatly from access to inet_pton/inet_ntop()
-    function calls in Python's socket module. Sadly, they aren't available so
-    we'll have to put up with the pure-Python implementation here (for now at
-    least).
+    Implements the operations that can be performed on an IPv6 network address
+    in accordance with RFC 4291.
     """
     def __init__(self):
         """Constructor."""
-        super(self.__class__, self).__init__(addr_type=AT_INET6,
+        super(IPv6Strategy, self).__init__(addr_type=AT_INET6,
             width=128, word_size=16, word_fmt='%x', word_sep=':')
 
     def valid_str(self, addr):
@@ -537,165 +492,37 @@ class IPv6Strategy(AddrStrategy):
         @return: C{True} if IPv6 network address string is valid, C{False}
             otherwise.
         """
-        #TODO: Reduce the length of this method ...
-        if not isinstance(addr, (str, unicode)):
+        if addr == '':
+            raise AddrFormatError('Empty strings are not supported!')
+
+        try:
+            _inet_pton(_AF_INET6, addr)
+        except _socket.error:
+            return False
+        except TypeError:
+            return False
+        except ValueError:
             return False
-
-        if '::' in addr:
-            #   IPv6 compact mode.
-            try:
-                prefix, suffix = addr.split('::')
-            except ValueError:
-                return False
-
-            l_prefix = []
-            l_suffix = []
-
-            if prefix != '':
-                l_prefix = prefix.split(':')
-
-            if suffix != '':
-                l_suffix = suffix.split(':')
-
-            #   IPv6 compact IPv4 compatibility mode.
-            if len(l_suffix) and '.' in l_suffix[-1]:
-                ipv4_str = l_suffix[-1]
-                if ST_IPV4.valid_str(ipv4_str):
-                    ipv4_int = ST_IPV4.str_to_int(ipv4_str)
-                    ipv4_words = ST_IPV4.int_to_words(ipv4_int)
-                    l_suffix.pop()
-                    l_suffix.append(
-                        ''.join(["%.2x" % i for i in ipv4_words[0:2]]))
-                    l_suffix.append(
-                        ''.join(["%.2x" % i for i in ipv4_words[2:4]]))
-
-            token_count = len(l_prefix) + len(l_suffix)
-
-            if not 0 <= token_count <= self.num_words - 1:
-                return False
-
-            try:
-                for token in l_prefix + l_suffix:
-                    word = int(token, 16)
-                    if not 0 <= word <= self.max_word:
-                        return False
-            except ValueError:
-                return False
-        else:
-            #   IPv6 verbose mode.
-            if ':' in addr:
-                tokens = addr.split(':')
-
-                if '.' in addr:
-                    ipv6_prefix = tokens[:-1]
-                    if ipv6_prefix[:-1] != ['0', '0', '0', '0', '0']:
-                        return False
-                    if ipv6_prefix[-1].lower() not in ('0', 'ffff'):
-                        return False
-                    #   IPv6 verbose IPv4 compatibility mode.
-                    if len(tokens) != (self.num_words - 1):
-                        return False
-                    ipv4_str = tokens[-1]
-                    if ST_IPV4.valid_str(ipv4_str):
-                        ipv4_int = ST_IPV4.str_to_int(ipv4_str)
-                        ipv4_words = ST_IPV4.int_to_words(ipv4_int)
-                        tokens.pop()
-                        tokens.append(
-                            ''.join(["%.2x" % i for i in ipv4_words[0:2]]))
-                        tokens.append(
-                            ''.join(["%.2x" % i for i in ipv4_words[2:4]]))
-                else:
-                    #   IPv6 verbose mode.
-                    if len(tokens) != self.num_words:
-                        return False
-                try:
-                    for token in tokens:
-                        word = int(token, 16)
-                        if not 0 <= word <= self.max_word:
-                            return False
-                except ValueError:
-                    return False
-            else:
-                return False
-
         return True
 
     def str_to_int(self, addr):
         """
         @param addr: An IPv6 address in string form.
 
-        @return: The equivalent network byte order integer for a given IPv6
-            address.
+        @return: The equivalent unsigned integer for a given IPv6 address.
         """
-        if not self.valid_str(addr):
-            raise ValueError("'%s' is an invalid IPv6 address!" % addr)
-
-        values = []
-
-        if addr == '::':
-            #   Unspecified address.
-            return 0
-        elif '::' in addr:
-            #   Abbreviated form IPv6 address.
-            prefix, suffix = addr.split('::')
-
-            if prefix == '':
-                l_prefix = ['0']
-            else:
-                l_prefix = prefix.split(':')
-
-            if suffix == '':
-                l_suffix = ['0']
-            else:
-                l_suffix = suffix.split(':')
-
-            #   Check for IPv4 compatibility address form.
-            if len(l_suffix) and '.' in l_suffix[-1]:
-                if len(l_suffix) > 2:
-                    return False
-                if len(l_suffix) == 2 and l_suffix[0].lower() != 'ffff':
-                    return False
-
-                ipv4_str = l_suffix[-1]
-                if ST_IPV4.valid_str(ipv4_str):
-                    ipv4_int = ST_IPV4.str_to_int(ipv4_str)
-                    ipv4_words = ST_IPV4.int_to_words(ipv4_int)
-                    l_suffix.pop()
-                    l_suffix.append(
-                        ''.join(["%.2x" % i for i in ipv4_words[0:2]]))
-                    l_suffix.append(
-                        ''.join(["%.2x" % i for i in ipv4_words[2:4]]))
-
-            gap_size = 8 - ( len(l_prefix) + len(l_suffix) )
-
-            values = ["%04x" % int(i, 16) for i in l_prefix] \
-                   + ['0000' for i in range(gap_size)] \
-                   + ["%04x" % int(i, 16) for i in l_suffix]
-        else:
-            #   Verbose form IPv6 address.
-            if '.' in addr:
-                #   IPv4 compatiblility mode.
-                tokens = addr.split(':')
-                ipv4_str = tokens[-1]
-                if ST_IPV4.valid_str(ipv4_str):
-                    ipv4_int = ST_IPV4.str_to_int(ipv4_str)
-                    ipv4_words = ST_IPV4.int_to_words(ipv4_int)
-                    tokens.pop()
-                    tokens.append(''.join(["%.2x" % i for i in ipv4_words[0:2]]))
-                    tokens.append(''.join(["%.2x" % i for i in ipv4_words[2:4]]))
-
-                values = ["%04x" % int(i, 16) for i in tokens]
-            else:
-                #   non IPv4 compatiblility mode.
-                values = ["%04x" % int(i, 16) for i in addr.split(':')]
-
-        value = int(''.join(values), 16)
-
-        return value
+        if addr == '':
+            raise AddrFormatError('Empty strings are not supported!')
+        try:
+            packed_int = _inet_pton(_AF_INET6, addr)
+            return self.packed_to_int(packed_int)
+        except Exception, e:
+            raise AddrFormatError('%r is not a valid IPv6 address string!' \
+                % addr)
 
     def int_to_str(self, int_val, compact=True, word_fmt=None):
         """
-        @param int_val: A network byte order integer.
+        @param int_val: An unsigned integer.
 
         @param compact: (optional) A boolean flag indicating if compact
             formatting should be used. If True, this method uses the '::'
@@ -703,92 +530,83 @@ class IPv6Strategy(AddrStrategy):
             of zero. Default: True
 
         @param word_fmt: (optional) The Python format string used to override
-            formatting for each word.
+            formatting for each word. Only applies when compact is False.
 
-        @return: The IPv6 string form equal to the network byte order integer
-        value provided.
+        @return: The IPv6 string form equal to the unsigned integer provided.
         """
-        #   Use basic parent class implementation if compact string form is
-        #   not required.
-        if not compact:
-            return super(self.__class__, self).int_to_str(int_val)
+        try:
+            packed_int = self.int_to_packed(int_val)
+            if compact:
+                #   Default return value.
+                return _inet_ntop(_AF_INET6, packed_int)
+            else:
+                #   Custom return value.
+                if word_fmt is None:
+                    word_fmt = self.word_fmt
+                words = list(_struct.unpack('>8H', packed_int))
+                tokens = [word_fmt % word for word in words]
+                return self.word_sep.join(tokens)
+        except Exception, e:
+            raise ValueError('%r is not a valid 128-bit unsigned integer!' \
+                % int_val)
 
-        _word_fmt = self.word_fmt
-        if word_fmt is not None:
-            _word_fmt = word_fmt
+    def int_to_packed(self, int_val):
+        """
+        @param int_val: the integer to be packed.
 
-        if not self.valid_int(int_val):
-            raise ValueError('%r is not a valid integer value supported ' \
-                'by this address type!' % int_val)
+        @return: a packed string that is equivalent to value represented by an
+        unsigned integer (in network byte order).
+        """
+        #   Here we've over-ridden the normal values to get the fastest
+        #   possible conversion speeds. Still quite slow versus IPv4 speed ups.
+        num_words = 4
+        word_size = 32
 
-        tokens = []
-        for i in range(self.num_words):
-            word = int_val & (2 ** self.word_size - 1)
-            tokens += [_word_fmt % word]
-            int_val >>= self.word_size
+        words = self.int_to_words(int_val, num_words, word_size)
 
-        tokens.reverse()
+        try:
+            fmt = '>%d%s'% (num_words, AddrStrategy.STRUCT_FORMATS[word_size])
+        except KeyError:
+            raise ValueError('unsupported word size: %d!' % word_size)
 
-        if compact == True:
-            new_tokens = []
+        return _struct.pack(fmt, *words)
 
-            positions = []
-            within_run = False
-            start_index = None
-            num_tokens = 0
+    def packed_to_int(self, packed_int):
+        """
+        @param packed_int: a packed string containing an unsigned integer.
+            It is assumed that string is packed in network byte order.
 
-            #   Discover all runs of zeros.
-            for idx, token in enumerate(tokens):
-                if token == '0':
-                    within_run = True
-                    if start_index is None:
-                        start_index = idx
-                    num_tokens += 1
-                else:
-                    if num_tokens > 1:
-                        positions.append((num_tokens, start_index))
-                    within_run = False
-                    start_index = None
-                    num_tokens = 0
+        @return: An unsigned integer that is equivalent to value of network
+            address represented by packed binary string.
+        """
+        #   Here we've over-ridden the normal values to get the fastest
+        #   conversion speeds. Still quite slow versus IPv4 speed ups.
+        num_words = 4
+        word_size = 32
 
-                new_tokens.append(token)
+        try:
+            fmt = '>%d%s'% (num_words, AddrStrategy.STRUCT_FORMATS[word_size])
+        except KeyError:
+            raise ValueError('unsupported word size: %d!' % word_size)
 
-            #   Store any position not saved before loop exit.
-            if num_tokens > 1:
-                positions.append((num_tokens, start_index))
+        words = list(_struct.unpack(fmt, packed_int))
 
-            #   Replace first longest run with an empty string.
-            if len(positions) != 0:
-                #   Locate longest, left-most run of zeros.
-                positions.sort(lambda x, y: cmp(x[1], y[1]))
-                best_position = positions[0]
-                for position in positions:
-                    if position[0] > best_position[0]:
-                        best_position = position
-                #   Replace chosen zero run.
-                (length, start_idx) = best_position
-                new_tokens = new_tokens[0:start_idx] + [''] + \
-                             new_tokens[start_idx+length:]
+        int_val = 0
+        for i, num in enumerate(reversed(words)):
+            word = num
+            word = word << word_size * i
+            int_val = int_val | word
 
-                #   Add start and end blanks so join creates '::'.
-                if new_tokens[0] == '':
-                    new_tokens.insert(0, '')
-
-                if new_tokens[-1] == '':
-                    new_tokens.append('')
-
-            tokens = new_tokens
-
-        return self.word_sep.join(tokens)
+        return int_val
 
     def int_to_arpa(self, int_val):
         """
-        @param int_val: A network byte order integer.
+        @param int_val: An unsigned integer.
 
         @return: The reverse DNS lookup for an IPv6 address in network byte
             order integer form.
         """
-        addr = self.int_to_str(int_val, word_fmt='%04x')
+        addr = self.int_to_str(int_val, compact=False, word_fmt='%.4x')
         tokens = list(addr.replace(':', ''))
         tokens.reverse()
         #   We won't support ip6.int here - see RFC 3152 for details.
@@ -828,7 +646,7 @@ class EUI48Strategy(AddrStrategy):
     #   equivalent.
     RE_MAC_FORMATS = [_re.compile(_, _re.IGNORECASE) for _ in RE_MAC_FORMATS]
 
-    def __init__(self, word_fmt='%02x', word_sep='-', uppercase=True):
+    def __init__(self, word_fmt='%.2X', word_sep='-'):
         """
         Constructor.
 
@@ -837,13 +655,10 @@ class EUI48Strategy(AddrStrategy):
 
         @param word_fmt: format string for each hextet.
             (Default: '%02x')
-
-        @param uppercase: return uppercase MAC/EUI-48 addresses.
-            (Default: True)
         """
         super(self.__class__, self).__init__(addr_type=AT_LINK, width=48,
-              word_size=8, word_fmt=word_fmt, word_sep=word_sep,
-              uppercase=uppercase)
+              word_size=8, word_fmt=word_fmt, word_sep=word_sep)
+
 
     def reset(self):
         """
@@ -856,10 +671,9 @@ class EUI48Strategy(AddrStrategy):
         self.num_words = self.width / self.word_size
         self.max_word = 2 ** self.word_size - 1
         self.word_sep = '-'
-        self.word_fmt  = '%02x'
+        self.word_fmt  = '%.2X'
         self.word_base = 16
         self.addr_type = AT_LINK
-        self.uppercase = True
 
     def valid_str(self, addr):
         """
@@ -880,8 +694,8 @@ class EUI48Strategy(AddrStrategy):
         """
         @param addr: An EUI-48 or MAC address in string form.
 
-        @return: A network byte order integer that is equivalent to value
-            represented by EUI-48/MAC string address.
+        @return: An unsigned integer that is equivalent to value represented
+            by EUI-48/MAC string address.
         """
         words = []
         if isinstance(addr, (str, unicode)):
@@ -905,13 +719,13 @@ class EUI48Strategy(AddrStrategy):
 
         if len(words) == 6:
             #   2 bytes x 6 (UNIX, Windows, EUI-48)
-            int_val = int(''.join(['%02x' % int(w, 16) for w in words]), 16)
+            int_val = int(''.join(['%.2x' % int(w, 16) for w in words]), 16)
         elif len(words) == 3:
             #   4 bytes x 3 (Cisco)
-            int_val = int(''.join(['%04x' % int(w, 16) for w in words]), 16)
+            int_val = int(''.join(['%.4x' % int(w, 16) for w in words]), 16)
         elif len(words) == 2:
             #   6 bytes x 2 (PostgreSQL)
-            int_val = int(''.join(['%06x' % int(w, 16) for w in words]), 16)
+            int_val = int(''.join(['%.6x' % int(w, 16) for w in words]), 16)
         elif len(words) == 1:
             #   12 bytes (bare, no delimiters)
             int_val = int('%012x' % int(words[0], 16), 16)
@@ -923,7 +737,7 @@ class EUI48Strategy(AddrStrategy):
 
     def int_to_str(self, int_val, word_sep=None, word_fmt=None):
         """
-        @param int_val: A network byte order integer.
+        @param int_val: An unsigned integer.
 
         @param word_sep: (optional) The separator used between words in an
             address string.
@@ -932,7 +746,7 @@ class EUI48Strategy(AddrStrategy):
             each word of address.
 
         @return: A MAC address in string form that is equivalent to value
-            represented by a network byte order integer.
+            represented by an unsigned integer.
         """
         _word_sep = self.word_sep
         if word_sep is not None:
@@ -946,32 +760,24 @@ class EUI48Strategy(AddrStrategy):
         tokens = [_word_fmt % i for i in words]
         addr = _word_sep.join(tokens)
 
-        if self.uppercase:
-            return addr.upper()
-        else:
-            return addr.lower()
+        return addr
 
 #-----------------------------------------------------------------------------
 #   Shared strategy objects for supported address types.
 #-----------------------------------------------------------------------------
 
 #: A shared strategy object supporting all operations on IPv4 addresses.
-ST_IPV4 = None
-#   Use the right strategy class dependent upon Python implementation (work-
-#   around for various Python bugs).
-if USE_IPV4_OPT is True:
-    ST_IPV4 = IPv4StrategyOpt()
-else:
-    ST_IPV4 = IPv4StrategyStd()
+ST_IPV4 = IPv4Strategy()
 
 #: A shared strategy object supporting all operations on IPv6 addresses.
-ST_IPV6  = IPv6Strategy()
+ST_IPV6 = IPv6Strategy()
+
 #: A shared strategy object supporting all operations on EUI-48/MAC addresses.
 ST_EUI48 = EUI48Strategy()
 
 #: A shared strategy object supporting all operations on EUI-64 addresses.
 ST_EUI64 = AddrStrategy(addr_type=AT_EUI64, width=64, word_size=8, \
-                         word_fmt='%02x', word_sep='-', uppercase=True)
+                         word_fmt='%.2X', word_sep='-')
 
 #-----------------------------------------------------------------------------
 if __name__ == '__main__':
